@@ -61,7 +61,7 @@ static void __exit mod_exit(void)
 	packet_deinit_data_caches();
 #endif
 	ratelimiter_module_deinit();
-	pr_debug("WireGuard has been unloaded\n");
+	pr_debug("WireGuard unloaded\n");
 }
 
 module_init(mod_init);
@@ -69,4 +69,5 @@ module_exit(mod_exit);
 MODULE_LICENSE("GPL v2");
 MODULE_DESCRIPTION("Fast, secure, and modern VPN tunnel");
 MODULE_AUTHOR("Jason A. Donenfeld <Jason@zx2c4.com>");
+MODULE_VERSION(WIREGUARD_VERSION);
 MODULE_ALIAS_RTNL_LINK(KBUILD_MODNAME);
